#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <signal.h>
/* for Solaris */
#include <curses.h>
/* for Linux */
/* 
#include <ncurses.h>
*/
#include <stdarg.h>
#include <time.h>
#include <unistd.h>
#include "yahtzee.h"


/*
 * (c)1992 by orest zborowski
 */

static char *header = YAHTZEE_HEADER;

extern errno;

#define MIN_NUM_ROLLS 2
#define DEFAULT_NUM_ROLLS 3
#define MAX_NUM_ROLLS 4
int dice_changed[5];			/* SBI_TRUE or SBI_FALSE */
int position_roll;			/* where the roll will go */

int NUM_ROLLS=DEFAULT_NUM_ROLLS;
char SCOREFNAME[80];
char LOWSCOREFNAME[80];

/* statistics about dice played */
int stat_dice[6]={0};
int stat_total=0;

static char *upper_headers[NUM_UPPER] =
{HEADER_A, HEADER_B, HEADER_C, HEADER_D, HEADER_E, HEADER_F};

static char *lower_headers[NUM_LOWER] =
{HEADER_G, HEADER_H, HEADER_I, HEADER_J, HEADER_K, HEADER_L, HEADER_M};

int dice_values[5];

static char *dice[6][4] =
{
  {"",
  "     ",
  "  o  ",
  "     "},

  {"",
  "o    ",
  "     ",
  "    o"},

  {"",
  "o    ",
  "  o  ",
  "    o"},

  {"",
  "o   o",
  "     ",
  "o   o"},

  {"",
  "o   o",
  "  o  ",
  "o   o"},

  {"",
  "o   o",
  "o   o",
  "o   o"}
};

Player players[MAX_NUMBER_OF_PLAYERS];

int num_players=0;

static int longest_header;

int numlines;

void init (void)
{
  int i;
  int j;

  sprintf(SCOREFNAME, "%s%d%s", SCOREFANTE, NUM_ROLLS, SCOREFPOST);
  sprintf(LOWSCOREFNAME, "%s%d%s", LOWSCOREFANTE, NUM_ROLLS, LOWSCOREFPOST);

  for (i = 0; i < MAX_NUMBER_OF_PLAYERS; ++i)
    {
      players[i].finished = 0;
      players[i].comp = 0;

      for (j = 0; j < NUM_FIELDS; ++j)
	{
	  players[i].score[j] = 0;
	  players[i].used[j] = 0;
	}
    }

  for (i = 0; i < NUM_UPPER; ++i)
    if (strlen (upper_headers[i]) > longest_header)
      longest_header = strlen (upper_headers[i]);

  for (i = 0; i < NUM_LOWER; ++i)
    if (strlen (lower_headers[i]) > longest_header)
      longest_header = strlen (lower_headers[i]);

  srand (getpid ());
}

static void 
fill_box (int x, int y, int h, int w)
{
  int dx, dy;
  for (dy = y; dy < y + h; dy++)
    {
      for (dx = x; dx < x + w; dx++)
	{
	  mvaddch (dx, dy, ' ');
	}
    }
}


/*
   **   SCREEN ORGANIZATION:
   **           line 0:         version header
   **           line 1:         -----------
   **           line 2:         edit window
   **           line 3:         -----------
   **           line 4:         player names
   **           line 5-10       upper bank
   **           line 11:        upper total
   **           line 12:        Bonus
   **           line 13-19:     lower bank
   **           line 20:        lower total
   **           line 21:        total
 */

void setup_screen (void)
{
  int i;

  initscr ();
  if (LINES < 25)
    sbi_abort (TERMINAL_TOO_SMALL);
  numlines = LINES;

  if (has_colors ())
    {
      start_color ();
      init_pair (COLOR_WHITE, COLOR_BLACK, COLOR_WHITE);
      init_pair (COLOR_GREEN, COLOR_BLACK, COLOR_GREEN);
      init_pair (COLOR_YELLOW, COLOR_WHITE, COLOR_GREEN);
      init_pair (COLOR_RED, COLOR_WHITE, COLOR_RED);
    }
  clear ();
  wattron (stdscr, COLOR_PAIR (COLOR_GREEN));
  fill_box (0, 0, COLS, LINES);
  mvaddstr (0, 9, header);
  move (1, 9);
  for (i = 9; i < COLS; ++i)
    addch (ACS_HLINE);
  refresh ();
}

void yend (void)
{
  attron (0);
  move (0, 0);
  clear ();
  refresh ();
  endwin ();
}

void sbi_abort (char *msg)
{
  yend ();
  putchar ('\n');
  printf (msg);
  putchar ('\n');
  exit (1);
}

void say (char *fmt,...)
{
  va_list ap;
  char buf[200];

  va_start (ap, fmt);
  vsprintf (buf, fmt, ap);
  va_end (ap);

  fill_box (2, 10, COLS - 10, 1);
  mvaddstr (2, 10, buf);
  refresh ();
}

void sbi_draw_position (char position, int so) {
  int i=position-'a';

  if ( (position<'a') || (position>'m') ) sbi_abort (BAD_POSITION);

  if (so==SBI_TRUE) {
    attroff (COLOR_PAIR (COLOR_WHITE));
    attron (A_STANDOUT);
  }
  else
    attron (COLOR_PAIR (COLOR_GREEN));

  if (i>=NUM_UPPER) {
    i -= NUM_UPPER;
    mvaddstr (i + 13, 9, lower_headers[i]);
  } else {
    mvaddstr (i + 5, 9, upper_headers[i]);
  }

  if (so==SBI_TRUE) attroff (A_STANDOUT);
  attron (COLOR_PAIR (COLOR_GREEN));
  refresh ();
} 

void sbi_draw_die (int num, int so) {
  int i,val;

  if ((num<1) || (num>5)) sbi_abort (BAD_DIE_NUMBER);
  val = dice_values[num-1];

  if (so==SBI_TRUE) {
    attroff (COLOR_PAIR (COLOR_BLACK));
    attron (A_STANDOUT);
  }
  else
    attron (COLOR_PAIR (COLOR_WHITE));

  for (i = 1; i < 4; ++i) {
    move (((num - 1) * 4) + i, 2);
    addstr (dice[val - 1][i]);
  }

  if (so==SBI_TRUE) attroff (A_STANDOUT);
  attron (COLOR_PAIR (COLOR_GREEN));
  refresh ();
}

void sbi_query_position
(int player, char *prompt, int *the_position)
{
  char c;
  int xpos;
  char position=0;

  xpos = 10 + strlen (prompt);
  if (player >= 0 && players[player].comp)	/* for the computer */
    {
      fill_box (2, 10, COLS - 10, 1);
      mvaddstr (2, 10, prompt);
      be_computer_position (player, the_position);
      refresh ();
      return;
    }

      cbreak ();
      noecho ();

      fill_box (2, 10, COLS - 10, 1);
      mvaddstr (2, 10, prompt);
      refresh ();

      /* read the keys until the first CR/LF */
      for (;;) {
	  c = getch ();
          if ( 
               ( ( (c==10) || (c==13) ) && (position != 0) )  /* CR/LF */
            || ((c>='a') && (c<='m')) /* swap a position */
            || ( c=='?') 
             ) {

	  /* LF or CR */
	  if (c == 10 || c == 13) {
            sbi_draw_position(position, SBI_FALSE);
            (*the_position) = (int)(position-'a');
	    break;
          } 

          if (c=='?') {
            if (position != 0) sbi_draw_position(position, SBI_FALSE);
            position = 0;
            be_computer_position (player, the_position);
            position = ('a'+(char)(*the_position));
            if ( (position>='a') && (position<='m') )
              sbi_draw_position(position, SBI_TRUE);
            else position = 0;
            move (2, 10+strlen(prompt)); 
          }

          /* regular die number */
	  if ( (c>='a') && (c<='m') ) {
            if (position != 0) sbi_draw_position(position, SBI_FALSE);
            position = c;
            sbi_draw_position(position, SBI_TRUE);
            move (2, 10+strlen(prompt)); 
          }

         }
        }

      echo ();
      nocbreak ();
}

void sbi_query_dice 
(int player, char *prompt, int *dice_roll, int *old_dice_roll)
{
  int i, j, die, flag=0; /* flag prevents two consecutive CR */
  char c;
  int old_reroll_flag={SBI_FALSE};

  for (i=0; i<5; i++) {
    dice_roll[i]=SBI_FALSE;
    if (old_dice_roll[i]==SBI_TRUE) old_reroll_flag=SBI_TRUE;
  }

  if (player >= 0 && players[player].comp)	/* for the computer */
    {
      if (old_reroll_flag) {
        fill_box (2, 10, COLS - 10, 1);
        mvaddstr (2, 10, prompt);
        be_computer_dice (player, dice_roll);
        refresh ();
      }
      return;
    }

      cbreak ();
      noecho ();

      fill_box (2, 10, COLS - 10, 1);
      mvaddstr (2, 10, prompt);
      refresh ();

      /* read the keys until the first CR/LF */
      for (;;) {
	  c = getch ();
          if ( 
               ( ( (c==10) || (c==13) ) && flag )  /* CR/LF */
            || (c=='0')            /* clear all dice */
            || (c=='9')            /* select all dice */
            || (c=='.')            /* redo last move */
            || (c=='?')            /* ask for advice */
            || ((c>='1') && (c<='5')) /* swap one die */
             ) {

	  /* LF or CR */
	  if (c == 10 || c == 13) {
	    break;
          } 

          flag = 1;

          if (c=='?') {
            be_computer_dice (player, dice_roll);
            for (j=0; j<5; j++) sbi_draw_die(j+1, dice_roll[j]);
            move (2, 10+strlen(prompt)); 
          }

          /* regular die number */
	  if ( (c>='1') && (c<='5') ) {
            die=c-'1';
            if (dice_roll[die]==SBI_TRUE) dice_roll[die]=SBI_FALSE;
              else dice_roll[die]=SBI_TRUE;
            sbi_draw_die(die+1, dice_roll[die]);
            move (2, 10+strlen(prompt)); 
          }

          /* redo (if no old redo, consider all is blank) */
          if (c=='.') for (j=0; j<5; j++) if (dice_roll[j]!=old_dice_roll[j]) {
            dice_roll[j]=old_dice_roll[j];
            sbi_draw_die(j+1, dice_roll[j]);
            move (2, 10+strlen(prompt)); 
          }

          /* clear all dice */
          if (c=='0') for (j=0; j<5; j++) if (dice_roll[j]!=SBI_FALSE) {
            dice_roll[j]=SBI_FALSE;
            sbi_draw_die(j+1, SBI_FALSE);
            move (2, 10+strlen(prompt)); 
          }

          /* select all dice */
          if (c=='9') for (j=0; j<5; j++) if (dice_roll[j]!=SBI_TRUE) {
            dice_roll[j]=SBI_TRUE;
            sbi_draw_die(j+1, SBI_TRUE);
            move (2, 10+strlen(prompt)); 
          }
         }
        }

      echo ();
      nocbreak ();
}

/* This returns a random number between 1 and 6 */
int raw_roll_dice (void) { return ((rand () % 6) + 1); }

/* This rolls a die and redraws it */
int roll_dice (int num)
{
  int val;
  int i;
  int j;

  if (num < 1 || num > 5)
    sbi_abort (BAD_DIE_NUMBER);

  attron (COLOR_PAIR (COLOR_GREEN));

  for (j = 0; j < 1; ++j)
    {
/*
   val = (random() % 6) + 1;
 */
      val = raw_roll_dice ();

      for (i = 0; i < 4; ++i)
	{
	  move (((num - 1) * 4) + i, 0);

	  if (i == 2)
	    printw ("%d", num);
	  else
	    addch (' ');
	  attroff (COLOR_PAIR (COLOR_GREEN));
	  attron (COLOR_PAIR (COLOR_WHITE));
	  if (i == 0)
	    {
	      if (num == 1)
		addch (ACS_ULCORNER);
	      else
		addch (ACS_LTEE);
	      addch (ACS_HLINE);
	      addch (ACS_HLINE);
	      addch (ACS_HLINE);
	      addch (ACS_HLINE);
	      addch (ACS_HLINE);
	      if (num == 1)
		addch (ACS_URCORNER);
	      else
		addch (ACS_RTEE);
	    }
	  else
	    {
	      addch (ACS_VLINE);
	      addstr (dice[val - 1][i]);
	      addch (ACS_VLINE);
	    }
	  attroff (COLOR_PAIR (COLOR_WHITE));
	  attron (COLOR_PAIR (COLOR_GREEN));
	}

      if (num == 5)		/* put the last +---+ tail */
	{
	  mvaddch (20, 0, ' ');
	  attroff (COLOR_PAIR (COLOR_GREEN));
	  attron (COLOR_PAIR (COLOR_WHITE));
	  addch (ACS_LLCORNER);
	  addch (ACS_HLINE);
	  addch (ACS_HLINE);
	  addch (ACS_HLINE);
	  addch (ACS_HLINE);
	  addch (ACS_HLINE);
	  addch (ACS_LRCORNER);
	  attroff (COLOR_PAIR (COLOR_WHITE));
	  attron (COLOR_PAIR (COLOR_GREEN));
	}
    }
  stat_dice[val-1]++;
  stat_total++;
  refresh ();
  return (val);
}

void print_stat() {
  int i;
  printf(STAT_HEADER);
  for(i=0; i<6; i++)
    printf(STAT_DIE, stat_dice[i], i+1, 
            ( (double)(100*stat_dice[i]) / (double)(stat_total) ), 37 );
  printf(STAT_FOOTER, stat_total);
}

int upper_total (int num)
{
  int val;
  int i;

  val = 0;

  for (i = 0; i < NUM_UPPER; ++i)
    val += players[num].score[i];

  return (val);
}

int
lower_total (int num)
{
  int val;
  int i;

  val = 0;

  for (i = 0; i < NUM_LOWER; ++i)
    val += players[num].score[i + NUM_UPPER];

  return (val);
}

int
total_score (int num)
{
  int upper_tot;
  int lower_tot;

  upper_tot = 0;
  lower_tot = 0;

  lower_tot = lower_total (num);
  upper_tot = upper_total (num);

  if (upper_tot >= 63)
    upper_tot += 35;

  return (upper_tot + lower_tot);
}

void
show_player (int num, int field)
{
  int i;
  int line;
  int upper_tot;
  int lower_tot;
  int xpos;

  xpos = 10 + longest_header + (num * MAX_NAME_LENGTH);

  for (i = 0; i < NUM_FIELDS; ++i)
    {
      if (i == field || field == -1)
	{
	  line = 5 + i;

	  if (i >= NUM_UPPER)
	    line += 2;

	  move (line, xpos);

	  if (players[num].used[i])
	    printw (" %4d", players[num].score[i]);

	  else
	    addstr ("     ");
	}
    }

  upper_tot = upper_total (num);
  lower_tot = lower_total (num);

  move (12, xpos);

  if (upper_tot >= 63)
    {
      printw ("+%4d", 35);
      upper_tot += 35;
    }
  else
    addstr ("    ");

  mvprintw (11, xpos, "(%4d)", upper_tot);
  mvprintw (20, xpos, "(%4d)", lower_tot);
  mvprintw (21, xpos, "[%4d]", upper_tot + lower_tot);

  refresh ();
}

void
setup_board (void)
{
  int i;
  int j;

  move (3, 9);
  for (i = 9; i < COLS; ++i)
    addch (ACS_HLINE);

  for (i = 0; i < NUM_UPPER; ++i)
    mvaddstr (i + 5, 9, upper_headers[i]);

  move (11, 9);
  attron (A_UNDERLINE | COLOR_PAIR (COLOR_YELLOW));
  addstr (UPPER_TOTAL);
  move (12, 9);
  addstr (BONUS);
  attroff (A_UNDERLINE | COLOR_PAIR (COLOR_YELLOW));
  attron (COLOR_PAIR (COLOR_GREEN));

  for (i = 0; i < NUM_LOWER; ++i)
    mvaddstr (i + 13, 9, lower_headers[i]);

  move (20, 9);
  attron (A_UNDERLINE | COLOR_PAIR (COLOR_YELLOW));
  addstr (LOWER_TOTAL);
  move (21, 9);
  addstr (GRAND_TOTAL);
  attroff (A_UNDERLINE | COLOR_PAIR (COLOR_YELLOW));
  attron (COLOR_PAIR (COLOR_GREEN));

  for (j = 0; j < num_players; ++j)
    {
      for (i = 4; i < 22; ++i)
	mvaddch (i, 9 + longest_header + (j * MAX_NAME_LENGTH),
		 ACS_VLINE);
    }

  for (i = 0; i < num_players; ++i)
    {
      mvaddstr (4, 10 + longest_header + (i * MAX_NAME_LENGTH),
		players[i].name);
      show_player (i, -1);
    }

  refresh ();
}

/* counts how many dice there are of a given value */
int count (int val)
{
  int i;
  int num;

  num = 0;

  for (i = 0; i < 5; ++i)
    if (dice_values[i] == val)
      ++num;

  return (num);
}

int find_n_of_a_kind (int n, int but_not)
{
  int i;

  for (i = 0; i < 5; ++i)
    {
      if (dice_values[i] == but_not)
	continue;

      if (count (dice_values[i]) >= n)
	return (dice_values[i]);
    }

  return (0);
}

int find_straight (int run, int notstart, int notrun)
{
  int i;
  int j;

  for (i = 1; i < 7; ++i)
    {
      if (i >= notstart && i < notstart + notrun)
	continue;

      for (j = 0; j < run; ++j)
	if (!count (i + j))
	  break;

      if (j == run)
	return (i);
    }

  return (0);
}

int find_yahtzee (void)
{
  int i;

  for (i = 1; i < 7; ++i)
    if (count (i) == 5)
      return (i);

  return (0);
}

int add_dice (void)
{
  int i;
  int val;

  val = 0;

  for (i = 0; i < 5; ++i)
    val += dice_values[i];

  return (val);
}

int showoff (int p, short so)
{
  move (4, 10 + longest_header + (p * MAX_NAME_LENGTH));

  if (so)
    {
      attroff (COLOR_PAIR (COLOR_GREEN));
      attron (A_STANDOUT);
    }
  else
    attron (COLOR_PAIR (COLOR_GREEN));

  addstr (players[p].name);

  if (so)
    {
      attroff (A_STANDOUT);
      attron (COLOR_PAIR (COLOR_GREEN));
    }
  refresh ();
  return 0;
}

void
handle_play (int player)
{
  int i;
  int dice_roll[5];		/* tells which dice to roll */
  int old_dice_roll[5];
  int position_roll;		/* tells which position to put things in */
  char msg[80];
  int done;
  int dummy;
  int numroll;
  int reroll_flag;

  for (i=0; i<5; i++) old_dice_roll[i]=SBI_FALSE;

  if (players[player].finished)	/* all finished */
    return;

  showoff (player, 1);

  say (I_M_ROLLING_DICE_FOR_PLAYER, players[player].name);

  for (i = 1; i < 6; ++i)
    dice_values[i - 1] = roll_dice (i);

  reroll_flag = SBI_TRUE;
  for (numroll = 1; ((numroll<NUM_ROLLS) && (reroll_flag==SBI_TRUE)); ++numroll)
    {
      sprintf(msg, WHAT_DICE_TO_ROLL_AGAIN, numroll+1, NUM_ROLLS);
      sbi_query_dice (player, msg, dice_roll, old_dice_roll);

      reroll_flag = SBI_FALSE;
      for (i=0; i<5; i++) {
        old_dice_roll[i]=dice_roll[i];
        if (dice_roll[i]==SBI_TRUE) reroll_flag=SBI_TRUE;
      }

      for (i=0; i<5; i++) if (dice_roll[i]==SBI_TRUE)
         dice_values[i]=roll_dice(i+1);
    }

  sbi_query_position (player, WHERE_TO_PUT_THAT, &position_roll);
  done = 0;

  for (;;)
    {
      if (position_roll < 0 || position_roll >= (NUM_LOWER+NUM_UPPER) )
	{
	  sbi_query_position 
            (player, BAD_CHOICE_POSITION, &position_roll);
	  continue;
	}

      switch (position_roll)
	{
	case 0:
	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
	  if (players[player].used[position_roll])
	    {
	      sbi_query_position (player, USED_POSITION, &position_roll);

	      break;
	    }

	  players[player].used[position_roll] = 1;

	  players[player].score[position_roll] =
	    count (position_roll + 1) * (position_roll + 1);

	  done = 1;

	  show_player (player, position_roll);

	  break;

	case 6:
	  if (players[player].used[position_roll])
	    {
	      sbi_query_position (player, USED_POSITION, &position_roll);

	      break;
	    }

	  players[player].used[position_roll] = 1;

	  if (find_n_of_a_kind (3, 0))
	    players[player].score[position_roll] =
	      add_dice ();

	  show_player (player, position_roll);

	  done = 1;

	  break;

	case 7:
	  if (players[player].used[position_roll])
	    {
	      sbi_query_position (player, USED_POSITION, &position_roll);

	      break;
	    }

	  players[player].used[position_roll] = 1;

	  if (find_n_of_a_kind (4, 0))
	    players[player].score[position_roll] =
	      add_dice ();

	  show_player (player, position_roll);

	  done = 1;

	  break;

	case 8:
	  if (players[player].used[position_roll])
	    {
	      sbi_query_position (player, USED_POSITION, &position_roll);

	      break;
	    }

	  players[player].used[position_roll] = 1;

	  dummy = find_n_of_a_kind (3, 0);

	  if (dummy != 0)
	    {
	      if (find_n_of_a_kind (2, dummy))
		players[player].score[position_roll] =
		  25;
	    }

	  show_player (player, position_roll);

	  done = 1;

	  break;

	case 9:
	  if (players[player].used[position_roll])
	    {
	      sbi_query_position (player, USED_POSITION, &position_roll);

	      break;
	    }

	  players[player].used[position_roll] = 1;

	  if (find_straight (4, 0, 0))
	    players[player].score[position_roll] = 30;

	  show_player (player, position_roll);

	  done = 1;

	  break;

	case 10:
	  if (players[player].used[position_roll])
	    {
	      sbi_query_position (player, USED_POSITION, &position_roll);

	      break;
	    }

	  players[player].used[position_roll] = 1;

	  if (find_straight (5, 0, 0))
	    players[player].score[position_roll] = 40;

	  show_player (player, position_roll);

	  done = 1;

	  break;

	case 11:
/*
   **   if the player scratched, the score for that position_roll will be 0.
   **   in that case, we don't allow it to be used any more.
 */
	  if ((players[player].score[position_roll] == 0 ||
	       !find_yahtzee ()) &&
	      players[player].used[position_roll] == 1)
	    {
	      sbi_query_position (player, SCRATCH_YAHTZEE, &position_roll);

	      break;
	    }

	  if (find_yahtzee ())
	    players[player].score[position_roll] += 50;

	  players[player].used[position_roll] = 1;

	  show_player (player, position_roll);

	  done = 1;

	  break;

	case 12:
	  if (players[player].used[position_roll])
	    {
	      sbi_query_position (player, USED_POSITION, &position_roll);

	      break;
	    }

	  players[player].used[position_roll] = 1;

	  players[player].score[position_roll] = add_dice ();

	  show_player (player, position_roll);

	  done = 1;

	  break;
	}

      if (done)
	break;
    }

  showoff (player, 0);

  for (i = 0; i < NUM_FIELDS; ++i)
    if (!players[player].used[i])
      return;

  players[player].finished = 1;
}

void
play (void)
{
  int i;
  int topscore;
  int winner;

  for (;;)
    {
      for (i = 0; i < num_players; ++i)
	handle_play (i);

      for (i = 0; i < num_players; ++i)
	if (!players[i].finished)
	  break;

      if (i == num_players)
	break;
    }

  topscore = -1;

  for (i = 0; i < num_players; ++i)
    if (total_score (i) > topscore)
      {
	topscore = total_score (i);

	winner = i;
      }

  say ("Le gagnant est %s", players[winner].name);
}

#define L_LOCK 0
#define L_UNLOCK 1

void
lock (char *fname, int type)
{
  char lockfile[200];
  struct stat statbuf;
  int i;
  FILE *fp;

  strcpy (lockfile, fname);

  strcat (lockfile, ".L");

  if (type == L_LOCK)
    {
      for (i = 1;; ++i)
	{
	  stat (lockfile, &statbuf);

	  if (errno == ENOENT)
	    break;

	  say (WAITING_FOR_LOCK, i);

	  sleep (1);
	}

      fp = fopen (lockfile, "w");

      fclose (fp);
    }

  else
    {
      unlink (lockfile);
    }
}

static int
write_score (FILE * fp, char *name, char *date, int score)
{
  return fprintf (fp, "%s\001%s\001%d\n", name, date, score);
}

static int
read_score (FILE * fp, char *name, char *date, int *score)
{
  char buf[200];
  char *sb, *se;

  if (!fgets (buf, sizeof (buf), fp))
    return -1;
  if (!(se = strchr (buf, '\001')))
    return -1;
  *se++ = '\0';
  strcpy (name, buf);
  if ((sb = se) > buf + sizeof (buf))
    return -1;
  if (!(se = strchr (sb, '\001')))
    return -1;
  *se++ = '\0';
  strcpy (date, sb);
  if ((sb = se) > buf + sizeof (buf))
    return -1;
  if (!(se = strchr (sb, '\n')))
    return -1;
  *se = '\0';
  *score = atoi (sb);
  return 0;
}

/*
   **   we keep track of the bottom nnn persons.
 */
void
update_bottom_scorefile (void)
{
  FILE *fp;
  FILE *tp;
  char scorefile[200];
  char tmpfile[100];
  int numtop;
  int nump;
  int j;
  char name[20];
  char date[30];
  char *curdate;
  int score;
  int topscore;
  int tmptop;
  int num_computers = 0;
  long clock;
  char scall[100];

  for (j = 0; j < num_players; ++j)
    if (players[j].comp == 1)
      num_computers++;

  if ((num_players - num_computers) > 0) 
    {

      sprintf (tmpfile, "/tmp/y.%d.%d", (int)(getpid()), (int)(getuid()));
      sprintf (scorefile, "%s/%s", SCOREDIR, LOWSCOREFNAME);

      clock = time (0);

      curdate = (char *) ctime (&clock);

      if (strchr (curdate, '\n'))
	*strchr (curdate, '\n') = '\0';

      if ((tp = fopen (tmpfile, "w")) == NULL)
	{
	  say (CAN_T_UPDATE_SCORE_FILE);
	  return;
	}

      lock (scorefile, L_LOCK);

      fp = fopen (scorefile, "r");

      for (j = 0; j < num_players; ++j)
	players[j].finished = -1;

      numtop = 0;
      nump = 0;
      topscore = MINUS_INFINITY;

      for (;;)
	{
/*
   **   get the next entry from the score file.  if there isn't any, then
   **   we set the score to beat to PLUS_INFINITY (everyone playing can beat it)
 */
	  if (fp == NULL || read_score (fp, name, date, &score))
	    score = PLUS_INFINITY;

/*
   **   now, we search through all players to find out which ones have scores
   **   higher than the one read (but less than topscore).  these will get
   **   saved before the read entry does. now, we only do this if all players
   **   haven't been accounted for.
 */
	  for (; nump != num_players;)
	    {
	      tmptop = PLUS_INFINITY;

	      for (j = 0; j < num_players; ++j)
		if (total_score (j) < tmptop &&
		    total_score (j) > topscore)
		  {
		    tmptop = total_score (j);
		  }

	      if (tmptop == PLUS_INFINITY)	/* everybody better */
		break;

	      if (tmptop < score)
		{
		  topscore = tmptop;

		  for (j = 0; j < num_players; ++j)
		    if (total_score (j) == tmptop)
		      {
			if (numtop >= NUM_TOP_PLAYERS)
			  break;

			if (players[j].comp != 1)
			  {
			    ++numtop;
			    write_score (tp,
					 players[j].name, curdate,
					 tmptop);
			  }
/* SBIREM 
 * Watch out not to increase numtop if nothing goes on to the temp file
 * or else there will miss entries...
 * This procedure is complicated, but it ends. Touch it as little as possible
 */

			players[j].finished = numtop - 1;

			++nump;

		      }
		}

	      else
		break;
	    }

	  if (score != PLUS_INFINITY && numtop < NUM_TOP_PLAYERS)
	    {
	      write_score (tp, name, date, score);
	      ++numtop;
	    }
/*
   **   if we processed all top slots or processed all players (and there
   **   was no score to beat), we stop.
 */
	  if (numtop == NUM_TOP_PLAYERS ||
	      (nump == num_players && score == PLUS_INFINITY))
	    break;
	}

      fclose (tp);
      if (fp)
	fclose (fp);

#ifdef HAS_RENAME
      if (rename (tmpfile, scorefile))
	{
	  say (RENAMING_FAILED);
	  unlink (tmpfile);
	}
#else
      sprintf (scall, "/bin/mv %s %s", tmpfile, scorefile);
      system (scall);
#endif

      lock (scorefile, L_UNLOCK);
    }
}

/*
   **   we keep track of the top nnn persons.
 */
void
update_top_scorefile (void)
{
  FILE *fp;
  FILE *tp;
  char scorefile[200];
  char tmpfile[100];
  int numtop;
  int nump;
  int j;
  char name[20];
  char date[30];
  char *curdate;
  int score;
  int topscore;
  int tmptop;
  int num_computers = 0;
  long clock;
  char scall[100];

  for (j = 0; j < num_players; ++j)
    if (players[j].comp == 1)
      num_computers++;

  if ((num_players - num_computers) > 0)
    {

      sprintf (tmpfile, "/tmp/y.%d.%d", (int)(getpid()), (int)(getuid()));
      sprintf (scorefile, "%s/%s", SCOREDIR, SCOREFNAME);

      clock = time (0);

      curdate = (char *) ctime (&clock);

      if (strchr (curdate, '\n'))
	*strchr (curdate, '\n') = '\0';

      if ((tp = fopen (tmpfile, "w")) == NULL)
	{
	  say (CAN_T_UPDATE_SCORE_FILE);
	  return;
	}

      lock (scorefile, L_LOCK);

      fp = fopen (scorefile, "r");

      for (j = 0; j < num_players; ++j)
	players[j].finished = -1;

      numtop = 0;
      nump = 0;
      topscore = PLUS_INFINITY;

      for (;;)
	{
/*
   **   get the next entry from the score file.  if there isn't any, then
   **   we set the score to beat to MINUS_INFINITY (everyone playing can beat it)
 */
	  if (fp == NULL || read_score (fp, name, date, &score))
	    score = MINUS_INFINITY;

/*
   **   now, we search through all players to find out which ones have scores
   **   higher than the one read (but less than topscore).  these will get
   **   saved before the read entry does. now, we only do this if all players
   **   haven't been accounted for.
 */
	  for (; nump != num_players;)
	    {
	      tmptop = MINUS_INFINITY;

	      for (j = 0; j < num_players; ++j)
		if (total_score (j) > tmptop &&
		    total_score (j) < topscore)
		  {
		    tmptop = total_score (j);
		  }

	      if (tmptop == MINUS_INFINITY)	/* everybody better */
		break;

	      if (tmptop > score)
		{
		  topscore = tmptop;

		  for (j = 0; j < num_players; ++j)
		    if (total_score (j) == tmptop)
		      {
			if (numtop >= NUM_TOP_PLAYERS)
			  break;

			if (players[j].comp != 1)
			  {
			    ++numtop;
			    write_score (tp,
					 players[j].name, curdate,
					 tmptop);
			  }
/* SBIREM 
 * Watch out not to increase numtop if nothing goes on to the temp file
 * or else there will miss entries...
 * This procedure is complicated, but it ends. Touch it as little as possible
 */

			players[j].finished = numtop - 1;

			++nump;

		      }
		}

	      else
		break;
	    }

	  if (score != MINUS_INFINITY && numtop < NUM_TOP_PLAYERS)
	    {
	      write_score (tp, name, date, score);
	      ++numtop;
	    }
/*
   **   if we processed all top slots or processed all players (and there
   **   was no score to beat), we stop.
 */
	  if (numtop == NUM_TOP_PLAYERS ||
	      (nump == num_players && score == MINUS_INFINITY))
	    break;
	}

      fclose (tp);
      if (fp)
	fclose (fp);

#ifdef HAS_RENAME
      if (rename (tmpfile, scorefile))
	{
	  say (RENAMING_FAILED);
	  unlink (tmpfile);
	}
#else
      sprintf (scall, "/bin/mv %s %s", tmpfile, scorefile);
      system (scall);
#endif

      lock (scorefile, L_UNLOCK);
    }
}

void
show_bottom_scores (void)
{
  FILE *fp;
  char scorefile[200];
  int i, j, k, score;
  char stuff[1024];
  char name[32], date[32];
  char msg[80];

  if (NUM_ROLLS==DEFAULT_NUM_ROLLS)
    sprintf(msg, WORST_SCORES_STANDARD_YAHTZEE, NUM_ROLLS);
  else
    sprintf(msg, WORST_SCORES_GENERIC_YAHTZEE, NUM_ROLLS);
  attron (COLOR_PAIR (COLOR_RED));
  fill_box (0, 0, COLS, LINES);
  attroff (COLOR_PAIR (COLOR_RED));
  attron (COLOR_PAIR (COLOR_WHITE));
  fill_box (0, 0, COLS, 1);
  move (0, (COLS - strlen (msg)) / 2);
  addstr (msg);
  attroff (COLOR_PAIR (COLOR_WHITE));

  sprintf (scorefile, "%s/%s", SCOREDIR, LOWSCOREFNAME);

  if ((fp = fopen (scorefile, "r")) == NULL)
    {
      yend ();
      printf (CAN_T_FIND_SCORE_FILE);
      return;
    }

  j = 0;
  attron (COLOR_PAIR (COLOR_RED));

  for (i = 0; i < NUM_TOP_PLAYERS; ++i)
    {
      if (read_score (fp, name, date, &score))
	break;

      if (j >= numlines - 4)
	{
	  move (23, (COLS - strlen (PRESS_ENTER)) / 2);
	  addstr (PRESS_ENTER);
	  refresh ();
	  getch ();
	  j = 0;
	}

      for (k = 0; k < num_players; ++k)
	if (players[k].finished == i)
	  break;
      fill_box (j + 3, 4, COLS - 4, 1);
      move (j + 3, 4);
      sprintf (stuff, "%3d : %-10s %s %d", i + 1, name, date, score);
      addstr (stuff);
      ++j;
    }

  fclose (fp);
  move (23, (COLS - strlen (PRESS_ENTER)) / 2);
  addstr (PRESS_ENTER);
  refresh ();
  getch ();
  yend ();
}

void
show_top_scores (void)
{
  FILE *fp;
  char scorefile[200];
  int i, j, k, score;
  char stuff[1024];
  char name[32], date[32];
  char msg[80];

  if (NUM_ROLLS==DEFAULT_NUM_ROLLS)
    sprintf(msg, BEST_SCORES_STANDARD_YAHTZEE, NUM_ROLLS);
  else
    sprintf(msg, BEST_SCORES_GENERIC_YAHTZEE, NUM_ROLLS);
  attron (COLOR_PAIR (COLOR_RED));
  fill_box (0, 0, COLS, LINES);
  attroff (COLOR_PAIR (COLOR_RED));
  attron (COLOR_PAIR (COLOR_WHITE));
  fill_box (0, 0, COLS, 1);
  move (0, (COLS - strlen (msg)) / 2);
  addstr (msg);
  attroff (COLOR_PAIR (COLOR_WHITE));

  sprintf (scorefile, "%s/%s", SCOREDIR, SCOREFNAME);

  if ((fp = fopen (scorefile, "r")) == NULL)
    {
      yend ();
      printf (CAN_T_FIND_SCORE_FILE);
      return;
    }

  j = 0;
  attron (COLOR_PAIR (COLOR_RED));

  for (i = 0; i < NUM_TOP_PLAYERS; ++i)
    {
      if (read_score (fp, name, date, &score))
	break;

      if (j >= numlines - 4)
	{
	  move (23, (COLS - strlen (PRESS_ENTER)) / 2);
	  addstr (PRESS_ENTER);
	  refresh ();
	  getch ();
	  j = 0;
	}

      for (k = 0; k < num_players; ++k)
	if (players[k].finished == i)
	  break;
      fill_box (j + 3, 4, COLS - 4, 1);
      move (j + 3, 4);
      sprintf (stuff, "%3d : %-10s %s %d", i + 1, name, date, score);
      addstr (stuff);
      ++j;
    }

  fclose (fp);
  move (23, (COLS - strlen (PRESS_ENTER)) / 2);
  addstr (PRESS_ENTER);
  refresh ();
  getch ();
  yend ();
}

void
calc_random (void)
{
  char nrollstr[10];
  int nroll;
  int table[NUM_FIELDS];
  int i;
  int j;

  printf (HOW_MANY_RANDOM_ROLLS);

  gets (nrollstr);
  nroll = atoi (nrollstr);

  printf ("Je calcule...\n");

  for (i = 0; i < NUM_FIELDS; ++i)
    table[i] = 0;

  for (i = 0; i < nroll; ++i)
    {
      for (j = 0; j < 5; ++j)
	dice_values[j] = raw_roll_dice ();

      for (j = 1; j <= 6; ++j)
	if (count (j) > 0)
	  ++table[j - 1];

      if (find_n_of_a_kind (3, 0))
	++table[6];

      if (find_n_of_a_kind (4, 0))
	++table[7];

      j = find_n_of_a_kind (3, 0);

      if (j != 0 && find_n_of_a_kind (2, j))
	++table[8];

      if (find_straight (4, 0, 0))
	++table[9];

      if (find_straight (5, 0, 0))
	++table[10];

      if (find_yahtzee ())
	++table[11];
    }

  printf ("%-25s %20s %20s\n", RANDOM_RESULT, RANDOM_TRIES, RANDOM_TOTAL);

  for (i = 0; i < NUM_FIELDS; ++i)
    {
      if (i < NUM_UPPER)
	printf ("%-35s", upper_headers[i]);

      else
	printf ("%-35s", lower_headers[i - NUM_UPPER]);

      printf (" %10d %20ld\n", table[i],
	      (long) (table[i] * 100) / nroll);
    }
}

void
signal_trap ()
{
  yend ();
  exit (0);
}

void set_signal_traps ()
{
  signal (SIGHUP, signal_trap);
  signal (SIGINT, signal_trap);
  signal (SIGQUIT, signal_trap);
}

void print_help(char *prog_name) {
printf (HELP_HEADER, prog_name);
printf (HELP_H_FLAG);
printf (HELP_S_FLAG);
printf (HELP_P_FLAG);
printf (HELP_C_FLAG);
printf (HELP_L_FLAG, MIN_NUM_ROLLS, MAX_NUM_ROLLS, DEFAULT_NUM_ROLLS);
printf (HELP_R_FLAG);
}

void main (int argc, char **argv)
{
  char num[10];
  int i;
  int num_computers = 0;
  short onlyshowscores = 0;
  char *prog_name = argv[0];
  int autostart = SBI_FALSE; /* set to SBI_TRUE if a player or a computer */
                             /* appears in the command line */

  init ();
  while (--argc > 0)
    {
      if ((*++argv)[0] == '-')
	{
	  switch ((*argv)[1])
	    {
            case 'p':
              if (num_players<MAX_NUMBER_OF_PLAYERS) {
                autostart = SBI_TRUE;
                for (i=2; (i<MAX_NAME_LENGTH+2) && ((*argv)[i]!=0); i++)
                  players[num_players].name[i-2] = (*argv)[i];
                players[num_players].name[i-2] = 0;
                if (players[num_players].name[0] == 0)
                  strcat(players[num_players].name, DEFAULT_NAME);
                num_players++;
              }
              break;

            case 'c':
              if (num_players<MAX_NUMBER_OF_PLAYERS) {
                autostart = SBI_TRUE;
                i=atoi((*argv)+2);
                if (i==0) i=1;
                if ( (num_players+i > MAX_NUMBER_OF_PLAYERS) )
                  i = MAX_NUMBER_OF_PLAYERS - num_players;
                for ( ; i>0; i--) {
                  players[num_players].comp = 1;
	          sprintf (players[num_players].name, COMPUTER_NAME,
                    num_computers + 'A');
                  num_computers++;
                  num_players++;
                }
              }
              break;

            case 'l':
              NUM_ROLLS = atoi((*argv)+2);
              if ( (NUM_ROLLS<MIN_NUM_ROLLS) || (NUM_ROLLS>MAX_NUM_ROLLS) )
                NUM_ROLLS = DEFAULT_NUM_ROLLS;
              sprintf(SCOREFNAME, "%s%d%s", 
                SCOREFANTE, NUM_ROLLS, SCOREFPOST);
              sprintf(LOWSCOREFNAME, "%s%d%s", 
                LOWSCOREFANTE, NUM_ROLLS, LOWSCOREFPOST);
              break;

	    case 's':
	      onlyshowscores = 1;
	      break;

	    case 'r':
	      calc_random ();
	      exit (0);

	    default:
              print_help(prog_name);
	      exit (0);
	    }
	}
    }

  if ( (!onlyshowscores) && (autostart==SBI_FALSE) )
    {
      printf (WELCOME_TO_YAHTZEE);


      do
	{
	  printf (HOW_MANY_HUMANS_PLAY, MAX_NUMBER_OF_PLAYERS);
	  fflush (stdout);

	  fgets (num, 10, stdin);

	  if (strchr (num, '\n'))
	    *strchr (num, '\n') = '\0';

	  num_players = atoi (num);

	  if (num_players == 0)
	    break;
	}
      while (num_players < 1 || num_players > MAX_NUMBER_OF_PLAYERS);

      for (i = 0; i < num_players; ++i)
	{
	  printf (WHAT_IS_NAME_OF_PLAYER, i + 1);
	  fflush (stdout);

	  fgets (players[i].name, MAX_NAME_LENGTH, stdin);

	  if (strchr (players[i].name, '\n'))
	    *strchr (players[i].name, '\n') = '\0';

          if (players[i].name[0] == 0)
            strcat(players[i].name, DEFAULT_NAME);
	}

      if (num_players == MAX_NUMBER_OF_PLAYERS)
	{
	  printf (COMPLAIN_CAN_T_PLAY);
	}

      else
	{
	  do
	    {
	      printf (HOW_MANY_COMPUTERS_PLAY, 
                      MAX_NUMBER_OF_PLAYERS - num_players);
	      fflush (stdout);

	      fgets (num, sizeof (num), stdin);

	      num_computers = atoi (num);
	    }
	  while (num_computers < 0 ||
		 num_players + num_computers > MAX_NUMBER_OF_PLAYERS);

	  for (i = 0; i < num_computers; ++i)
	    {
	      players[num_players].comp = 1;

	      sprintf (players[num_players].name, "Mr. %c",
		       i + 'A');

	      ++num_players;
	    }
	}

      if (num_players == 0) {
        printf (WHY_DID_YOU_RUN_THIS_THEN);
        exit (8);
      }
    }

  setup_screen ();

  set_signal_traps ();

  if (!onlyshowscores)
    {
      setup_board ();

      play ();

      update_top_scorefile ();
      update_bottom_scorefile ();
    }

  getch ();
  show_top_scores ();
  show_bottom_scores ();

  if (!onlyshowscores)
    print_stat();
  exit (0);
}
